#include "frontend.h"
#include <sys/stat.h>
#include "keydef.h"
#include <sys/types.h>
#include <fcntl.h>
#include <vga.h>
#include <vgagl.h>
#include <tiffio.h>


void
find_best_mode(int width, int height, char *best_mode) {
	
best_mode_temp = best_mode_256tail;
	while(best_mode_temp->preventry != NULL &&  (width < best_mode_temp->min_width &&
       	  height < best_mode_temp->min_height) ) {
		best_mode_temp = best_mode_temp->preventry;
	}

   	if(best_mode_temp->nextentry != NULL) { best_mode_temp = best_mode_temp->nextentry; }
    strcpy(best_mode, best_mode_temp->mode_name);
}


void
file_info(char *file_name, int length) {


	WINDOW *info;
	struct stat	file_name_info;
	FILE   *graphic_file_test;
	char   found_file_type[32], *line, best_mode[10], testline[64];
	int    c, image_width, image_height;
	int    header[450], i, j, store[3];
    TIFF   *tiff_graphic_file_test;

    /* gifs and targa use 2bit height and width while png, jpeg and 
        bitmaps use 4bit, png use 4bit rev, jpeg 2bit rev */
	#define UCH(x)	((int) (x))
	#define GET_2B(array,offset) ((unsigned int) UCH(array[offset]) + \
			       (((unsigned int) UCH(array[offset+1])) << 8))
	#define GET_REV_2B(array,offset)  ((((int) UCH(array[offset])) << 8) + \
                   ((int) UCH(array[offset+1])))
	#define GET_4B(array,offset)  ((int) UCH(array[offset]) + \
			       (((int) UCH(array[offset+1])) << 8) + \
			       (((int) UCH(array[offset+2])) << 16) + \
			       (((int) UCH(array[offset+3])) << 24))
	#define GET_REV_4B(array,offset)  (((int) UCH(array[offset]) << 24) + \
			       (((int) UCH(array[offset+1])) << 16) + \
			       (((int) UCH(array[offset+2])) << 8) + \
			       ((int) UCH(array[offset+3])))

	/* file file format */
	strcpy(found_file_type, "Text File"); /* default string type */
	if((graphic_file_test = fopen(file_name, "rb")) == NULL) { fprintf(dfd, "file empty\n"); }
	if((c = getc(graphic_file_test)) != EOF) {
		switch (c) {
#ifndef NO_GIF
			case 'G':
			    strcpy(found_file_type, "GIF");
				fseek(graphic_file_test, 0, 0);
        		for(i = 0; i < 12; i++) { header[i] = getc(graphic_file_test); }
				image_width  = (int) GET_2B(header,6);
				image_height = (int) GET_2B(header,8);
		    	break;
#endif
#ifndef NO_TARGA
			case 0x00:
			    strcpy(found_file_type, "TARGA");
				fseek(graphic_file_test, 0, 0);
				for(i = 0; i < 18; i++) { header[i] = getc(graphic_file_test); }
				image_width  = (int) GET_2B(header,12);
				image_height = (int) GET_2B(header,14);
                break;
#endif
#ifndef NO_BMP
			case 'B' :
				strcpy(found_file_type, "BITMAP");
				fseek(graphic_file_test, 0, 0);
				for(i = 0; i < 27; i++) { header[i] = getc(graphic_file_test); }
				image_width  = (int) GET_4B(header,18);
				image_height = (int) GET_4B(header,22);
				break;
#endif
#ifndef NO_JPEG
			case 0xff:
				strcpy(found_file_type, "JPEG");
                fseek(graphic_file_test, 0, 0);
        		for(i = 0; i < 450; i++) { header[i] = getc(graphic_file_test); }
                i = 25;
                while( header[i] != 255 || header[i + 1] != 192 ) i++;
            	i = i + 5;
				image_width  = (int) GET_REV_2B(header,i);
                i = i + 2;
				image_height = (int) GET_REV_2B(header,i);
				break;
#endif
#ifndef NO_PNG
            case 0x89 :
 				strcpy(found_file_type, "PNG");
				fseek(graphic_file_test, 0, 0);
        		for(i = 0; i < 25; i++) { header[i] = getc(graphic_file_test); }
				image_width  = (int) GET_REV_4B(header,16);
				image_height = (int) GET_REV_4B(header,20);
                break;
#endif
#ifndef NO_TIFF
            case 'I' :
 				strcpy(found_file_type, "TIFF");
                tiff_graphic_file_test = TIFFOpen(file_name, "r");
            	TIFFGetField(tiff_graphic_file_test, TIFFTAG_IMAGEWIDTH, &image_width);
                TIFFGetField(tiff_graphic_file_test, TIFFTAG_IMAGELENGTH, &image_height);
                TIFFClose(tiff_graphic_file_test);
                break;
#endif
#ifndef NO_PPM
            case 'P' :
                c = getc(graphic_file_test);
                if(c == '6') strcpy(found_file_type, "PPM");
                else strcpy(found_file_type, "PGM");
                do { c = getc(graphic_file_test); } while(c != '\n'); /* finnish first line */
                if(getc(graphic_file_test) == '#') { do { c = getc(graphic_file_test); } while(c != '\n'); }                            fgets(testline, 64, graphic_file_test); /* read info line */
                for(i = 0; i < 64; i++) if(testline[i] < '0' || testline[i] > '9') testline[i] = ':';
                for(i = 0; testline[i] < '0' || testline[i] > '9'; i++);
                for(j = 0; testline[i] != ':'; j++) store[j] = convert_to_int(testline[i++]);
                image_width = return_value(store, j);

                for(; testline[i] < '0' || testline[i] > '9'; i++);
                for(j = 0; testline[i] != ':'; j++) store[j] = convert_to_int(testline[i++]);
                image_height = return_value(store, j);
                break;   
#endif
			case '/' :
			case '#' :
				c = getc(graphic_file_test);
				switch(c) {
#ifndef NO_PIXMAP
					case '*' :
                        fgets(testline, 64, graphic_file_test); /* read info line */
                	    if(strncmp(" XPM */", testline, 7) == 0) { /* a test to exclude other c files */
                           strcpy(found_file_type, "PIXMAP");
                           fclose(graphic_file_test);
                           graphic_file_test = fopen(file_name, "r");

                        	for(; c != '"'; c = getc(graphic_file_test));
                        
                           image_width  =  convert_to_int(getc(graphic_file_test)) * 10;
                           image_width  += convert_to_int(getc(graphic_file_test));
                           c = getc(graphic_file_test);
                           image_height =  convert_to_int(getc(graphic_file_test)) * 10;
                           image_height += convert_to_int(getc(graphic_file_test));
                        }
                        else {
                           c = service_check(file_name);
                           if(c >= 40) strcpy(found_file_type, "Text File");
                           else strcpy(found_file_type, "Bin. Code");
                           image_width  = 0;
                           image_height = 0;
                           strcpy(best_mode, "Non-Aplc");
                        }
						break;
#endif
#ifndef NO_XBITMAP
					case 'd' : /* this should have a c file test but it is not pratical */
						strcpy(found_file_type, "XBITMAP");
                        fclose(graphic_file_test);
                        graphic_file_test = fopen(file_name, "r");

                    	for(; c != ' '; c = getc(graphic_file_test) );
                        c = getc(graphic_file_test);
                    	for(; c != ' '; c = getc(graphic_file_test) );
						image_width  = convert_to_int(getc(graphic_file_test)) * 10;
						image_width += convert_to_int(getc(graphic_file_test));

                    	for(; c != '#'; c = getc(graphic_file_test) );

                    	for(; c != ' '; c = getc(graphic_file_test) );
                        c = getc(graphic_file_test);
                    	for(; c != ' '; c = getc(graphic_file_test) );
						image_height  = convert_to_int(getc(graphic_file_test)) * 10;
						image_height += convert_to_int(getc(graphic_file_test));
                        break;
#endif
 				} /* end switch */
    		    break;
            default : 
                c = service_check(file_name);
                if(c >= 40) strcpy(found_file_type, "Text File");
                else strcpy(found_file_type, "Bin. Code");
                image_width  = 0;
                image_height = 0;
                strcpy(best_mode, "Non-Aplc");
                break;
		} /* end switch */
	} /* end if */
	fclose(graphic_file_test);
    
   	if(strncmp(best_mode, "Non-Aplc", 7) != 0) find_best_mode(image_height, image_width, best_mode);

	stat(file_name, &file_name_info);
	if(length < 10) { length = 10; }
   	info = newwin(8, length + 18, 5, 5 );
    wborder(info, ACS_VLINE, ACS_VLINE, ACS_HLINE, ACS_HLINE,
   	                 ACS_ULCORNER, ACS_URCORNER, ACS_LLCORNER, ACS_LRCORNER); 
   	mvwprintw(info, 1, 3, "File Name  : %s", file_name);

	mvwprintw(info, 2, 3, "File Type  : %s", found_file_type);
	mvwprintw(info, 3, 3, "File Size  : %dK", file_name_info.st_size/1000);
	mvwprintw(info, 4, 3, "File Height: %d", image_height);
	mvwprintw(info, 5, 3, "File Width : %d", image_width);
	mvwprintw(info, 6, 3, "Best Mode  : %s", best_mode);
    wrefresh(info);
	raw();
	noecho();
    i = wgetch(info);
	echo();
	noraw();
    if(i == KEY_CTRL_X) { runexit(DIS_ON); }
	delwin(info);
	redrawscreen();
}


void
addtostruct(int min_width2, int min_height2, int mode2,	int mode_width2, int mode_height2,
			int mode_depth2, char *mode_name2) {

	bmode *current;

	current = malloc(sizeof(bmode));

	current->min_width   = min_width2;
	current->min_height  = min_height2;
	current->mode 		 = mode2;
	current->mode_width  = mode_width2;
	current->mode_height = mode_height2;
	current->mode_depth	 = mode_depth2;
    current->mode_name   = mode_name2;
	current->nextentry   = NULL;	
    
    switch(current->mode_depth) {
       case 256:
          	if(best_mode_256head != NULL) {
               	best_mode_temp = best_mode_256head;
                while(best_mode_temp->nextentry != NULL) {
                	best_mode_temp = best_mode_temp->nextentry;
                }
                best_mode_temp->nextentry = current;
                current->preventry = best_mode_temp;
                best_mode_256tail = current;
            }
            else  {
                best_mode_256head = current;
                current->preventry = NULL;
                best_mode_256tail = current;
            }
            break;
       case 32768 :
          	if(best_mode_32Khead != NULL) {
               	best_mode_temp = best_mode_32Khead;
                while(best_mode_temp->nextentry != NULL) {
                	best_mode_temp = best_mode_temp->nextentry;
                }
                best_mode_temp->nextentry = current;
                current->preventry = best_mode_temp;
                best_mode_32Ktail = current;
            }
            else  {
                best_mode_32Khead = current;
                current->preventry = NULL;
                best_mode_32Ktail = current;
            }
            break;
       case 16777216 :
          	if(best_mode_16Mhead != NULL) {
               	best_mode_temp = best_mode_16Mhead;
                while(best_mode_temp->nextentry != NULL) {
                	best_mode_temp = best_mode_temp->nextentry;
                }
                best_mode_temp->nextentry = current;
                current->preventry = best_mode_temp;
                best_mode_16Mtail = current;
            }
            else  {
                best_mode_16Mhead = current;
                current->preventry = NULL;
                best_mode_16Mtail = current;
            }
            break;
    }
}


void
getmodeinfo() {

	int i, v = 0, j = 0;

    best_mode_256head	= NULL;
	best_mode_256tail	= NULL;
    best_mode_32Khead	= NULL;
	best_mode_32Ktail	= NULL;
    best_mode_16Mhead	= NULL;
	best_mode_16Mtail	= NULL;
	vga_init();

	for(i = 1; i <= GLASTMODE; i++) {
		if(vga_hasmode(i)) {
			switch (i) {
				case 5 :
					addtostruct(0, 0, 5, 320, 200, 256, "320x200");
					break;
				case 10 : 
					addtostruct(360, 480, 10, 640, 480, 256, "640x480");
					break;
				case 11 : 
					addtostruct(640, 480, 11, 800, 600, 256, "800x600");
					break;
				case 12 : 
					addtostruct(800, 600, 12, 1024, 768, 256, "1024x768");
					break;
                case 14 :
					addtostruct(1, 1, 14, 320, 200, 32768, "320x200");
					break;
				case 17 : 
					addtostruct(320, 200, 17, 640, 480, 32768, "640x480");
					break;
				case 20 : 
					addtostruct(640, 480, 20, 800, 600, 32768, "800x600");
					break;
				case 23 : 
					addtostruct(800, 600, 23, 1024, 768, 32768, "1024x768");
					break;
				case 33 : 
					addtostruct(1, 1, 33, 320, 200, 16777216, "320x200");
					break;
				case 34 : 
					addtostruct(320, 200, 34, 640, 480, 16777216, "640x480");
					break;
				case 35 : 
					addtostruct(640, 480, 35, 800, 600, 16777216, "800x600");
					break;
                default :
                    break;
			}
		}
	}
}
