#include "frontend.h"
#include "keydef.h"

#define MAX_WINDOWS 300 

DIRINFO dirwininfo;
opts    options;

LINKDIR *linked_directory;
LINKDIR *linked_directory_head;
LINKDIR *linked_directory_tail;
LINKDIR *linked_directory_temp;

bmode   *best_mode_256head;
bmode	*best_mode_256tail;
bmode   *best_mode_32Khead;
bmode	*best_mode_32Ktail;
bmode   *best_mode_16Mhead;
bmode	*best_mode_16Mtail;
bmode	*best_mode_temp;

FILE *dfd;

WINDOW *window[MAX_WINDOWS];

void
main(int argc, char *argv[]) {

char   *termvalue;

#ifdef VERSION
WINDOW *namewin;
#endif

  linked_directory_temp = malloc(sizeof(LINKDIR)); 
  linked_directory_head = NULL;

  termvalue = getenv("TERM");
  if(strncmp(termvalue, "xterm", 5) == 0) {
	printf("Sorry running in xterm can not run graphics exiting\n");
    exit(0);
  }
  initscr();
  clear();
  if(argc != 1) chdir(argv[1]); /* if a command argument exists attempt to change directory to it */
  getcwd(dirwininfo.dirpath, 127); /* get current directory */

  getmodeinfo();

#ifdef VERSION
  /* version window */
  namewin = newwin(7, 20, (LINES/2-4), (COLS-COLS/2-10) );
  mvwprintw(namewin, 1, 3, "Version 1.0b1");
  mvwprintw(namewin, 3, 6, "of mbgv");
  mvwprintw(namewin, 5, 3, "by Eric Gandt");
  mvwprintw(namewin, 7, 2, "gandt@wam.umd.edu");
  wborder(namewin, ACS_VLINE, ACS_VLINE, ACS_HLINE, ACS_HLINE, ACS_ULCORNER, ACS_URCORNER, ACS_LLCORNER, ACS_LRCORNER);
  wrefresh(namewin);
  raw();
  noecho();
  wgetch(namewin);
  echo();
  noraw();
  clear();
  refresh();
#endif
 
  start_up_windows();
}


void
start_up_windows() {

	int 	win_start_position, i;
	int 	win_end_position,   j;


	j = ((LINES-4)*3)+1;
	for (i=0; i < j; i++)
	{
		window[i]  = (WINDOW*)malloc(sizeof(WINDOW));
	}

	/* check position of first window on a no even screen 
	   it will need to be one larger than the rest */
	win_end_position = COLS/3;
	win_start_position = COLS-(COLS/3)*2;
	while(win_end_position != win_start_position) { 
		win_end_position++; 
	}

	/* set up windows */
	for(i = 0; i < LINES-4; i++) {
		window[i] 				= newwin(1, win_end_position, i+3, 0);
		window[i+(LINES-4)] 	= newwin(1, COLS/3, i+3, (COLS-(COLS/3)*2) );
		window[i+((LINES-4)*2)] = newwin(1, COLS/3, i+3, COLS-(COLS/3) );
	}

    dirwininfo.writeoffset = 0;

	/* read entries */
	dirwininfo.totalentries = readdirectories();

	linked_directory = linked_directory_head; 

dfd = fopen("debug", "w");
setlinebuf(dfd);
fprintf(dfd, "top\n");

	writeoutlist(dirwininfo.writeoffset);

	strcpy(dirwininfo.prevdir, dirwininfo.dirpath);

	dorefreshwindows();
	dealwithscroll();
}


void
dorefreshwindows() {
	int i;

	/* for every window redraw it */
	for(i = 0; i < LINES-4; i++) {
		wnoutrefresh(window[i] );
		wnoutrefresh(window[i+(LINES-4)] ); 
		wnoutrefresh(window[i+((LINES-4)*2)] ); 
	}
	doupdate();
}


void
dealwithscroll() {

    int  	j = 0, v = 0, x, y;
    int 	response, ret_val, enable, keep_force, use_force;
	char	dirpathtemp[128], dirpath[128];
    WINDOW	*entrywin;
        
    options.opt_greyscale  = OFF;
    options.opt_quantize   = OFF;
    options.opt_forcemode  = OFF;    
    options.opt_forcescale = ON;
    	
	/* set default options for viewing */
	dirwininfo.marked  = OFF; 

    entrywin = newwin(3, COLS, 0, 0);
	keypad(entrywin, TRUE);

    /* setup startup states */
    enable     = ON;
    ret_val    = ERROR_KEY;
    keep_force = OFF;
    use_force  = OFF;

	/* if there is no path prompt to set one */
    if(linked_directory_head != NULL) {
		linked_directory = linked_directory_head;
    }
    else {
		enternewdir("--");
    }
    
    while (1)  {

		/* set up window and write out usefull information */
   		wborder(entrywin, ACS_VLINE, ACS_VLINE, ACS_HLINE, ACS_HLINE,
    	               	  ACS_ULCORNER, ACS_URCORNER, ACS_LLCORNER,
						  ACS_LRCORNER); 
	
		if (linked_directory->itemlength > COLS/2-2) {
    	    mvwprintw(entrywin, 1, (COLS/2+2), "File: ");
		    for(j = 0; j < (COLS/2-2) && j < linked_directory->itemlength; j++) {
        	   	wprintw(entrywin, "%c", linked_directory->itemname[j]); 	
	    	}
		}
		else {
    	    mvwprintw(entrywin, 1, (COLS/2+2), "File: %s", linked_directory->itemname);
		}
		v = strlen(dirwininfo.dirpath);
		if(v > COLS/2-2 ) {
			mvwprintw(entrywin, 1, 1, "Dir: ^");
		    for (j = (v-COLS/2-2); j < v; j++) { wprintw(entrywin, "%c", dirwininfo.dirpath[j]); }
		}
		else {	mvwprintw(entrywin, 1, 1, "Dir: %s", dirwininfo.dirpath); }
		
		/* keeps the char for being echoed to the screen */
        if(enable == ON) {
           raw();
           noecho();
           response = wgetch(entrywin);
           echo();
           noraw();
        }

   		wrefresh(entrywin);
        
        if(use_force == OFF) options.opt_forcemode = keep_force;
        use_force = OFF;
        enable = ON;

   	switch (response) {
	   case KEY_DEL :
			if(linked_directory->itemtype != 'D') file_del();
            else { movedown(); }
			break;			
	   case KEY_INST :
			if(linked_directory->itemtype != 'D') file_copy();
			break;
	   case KEY_UP :
	   case KEY_K :
	   case KEY_k : /* up */
			moveup();
			break;
	   case KEY_DOWN :
	   case KEY_J :
	   case KEY_j :  /* down */
    	    movedown();
			break;
	   case KEY_NPAGE :
	   case KEY_RIGHT :
	   case KEY_L :
	   case KEY_l :  /* right */
			moverightwin();
			break;
	   case KEY_PPAGE :
	   case KEY_LEFT  :
	   case KEY_H :
	   case KEY_h :  /* left */
			moveleftwin();
			break;
	   case KEY_I :
	   case KEY_i : /* change view options */
			if(linked_directory->itemtype != 'D') file_info(linked_directory->itemname, linked_directory->itemlength);
			else movedown();
			break;
	   case	KEY_HELP : /* run help */
			runhelp();
			break; 
   	   case KEY_CTRL_X : /* quit the program */	
     		delwin(entrywin);
			runexit(DIS_ON); 
			break;
	   case KEY_ENTER : /* view current */
	   case KEY_N :
	   case KEY_n :
            if(linked_directory->itemtype == 'D') {
            	strcpy(dirpathtemp, linked_directory->itemname);
                dirpathtemp[linked_directory->itemlength] = '\0';
                getcwd(dirpath, 127); 
                strcat(dirpath, "/");
                strcat(dirpath, dirpathtemp);
                enternewdir(dirpath); 
            }
            else { 
               ret_val = read_graphics(linked_directory->itemname);
               switch(ret_val) {
                  case CTRL_X_KEY :
                     delwin(entrywin);
                     runexit(DIS_OFF);
                     break;
                  case RETURN_KEY :
                     enable = OFF;
                     movedown();
                     response = KEY_ENTER;
                     break;
                  case SPACE_KEY :
                     enable = OFF;
                     response = KEY_SPACE;
                     break;
                  case BACK_KEY :
		     if(linked_directory->preventry != NULL && linked_directory->preventry->itemtype == 'F') {
                          enable = OFF;
			  moveup();
                     	  response = KEY_ENTER;
		     }
		     else response = KEY_ESC;
                     break;
                  case ADD_KEY :
                  case SUB_KEY :
                     enable    = OFF;
                     use_force = ON;
                     response = KEY_ENTER;
                     break;                     
                  case ESC_KEY :
                  case ERROR_KEY :
                  default :
                     break;
               }
            }        
			break;
       case KEY_O :
       case KEY_o :
          	options_menu();
            break;
       case KEY_D :
       case KEY_d : /* change directories */
			enternewdir("--"); 
			break;
	   case KEY_R :
	   case KEY_r : /* refresh the screen  */
			redrawscreen();
			break;
	   case KEY_U :
	   case KEY_u : /* prevous directory */
			enternewdir(dirwininfo.prevdir); 
			break;
       case KEY_V :
       case KEY_v :
          	keep_force = options.opt_forcemode = runvideo();
          	break;
	   case KEY_SPACE : /* mark the current */
			markcurrent();
			movedown();
            if(ret_val == SPACE_KEY) {
               enable = OFF;
               response = KEY_ENTER;
            }
			break;
	   case KEY_HOME : /* goto first element of list */
			movehome();
			break;
	   case KEY_END :  /* goto last element of list */
			moveend();
			break;
	   default :
			break;
	}	    
	werase(entrywin);
    }	    
}


void
markcurrent() {

	if(linked_directory->itemname[linked_directory->itemlength-1] != '/') {
		if(linked_directory->itemmarked == ON) {
			linked_directory->itemmarked = OFF;
			mark_window(linked_directory->itemwindow, NORMAL);
		}
		else {
			linked_directory->itemmarked = ON;
			mark_window(linked_directory->itemwindow, BOLD);
		}
	}
}


void
redrawscreen() {

	/* clear and redraw the entire screen cleaning up left over crap */
	erase();
	refresh();
	writeoutlist(dirwininfo.writeoffset);	
	dorefreshwindows();
}

	
