/* include golobal libaries */
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <ncurses.h>

/* posible return values for set_it_all_running */
#define ERROR_KEY	0
#define RETURN_KEY  3
#define ESC_KEY	    5
#define CTRL_X_KEY  7 
#define SPACE_KEY	9
#define BACK_KEY 	11
#define ADD_KEY		1
#define SUB_KEY	   -1

/* defines for mark types */
#define BOLD 	1
#define NORMAL	3
#define REVERSE	5

/* define yes and no */
#define OFF 0
#define ON  1

/* define display options for menus on or off */
#define DIS_OFF 0
#define DIS_ON  1

/* defines down and up */
#define DOWN  -1
#define UP    1

typedef struct OPTS {
	int 	opt_greyscale;
	int 	opt_quantize;
	int 	opt_forcemode;
    int		opt_forcescale;
	} opts;

/* video mode info structure def */
typedef struct {
	int mode;
	int mode_width;
	int mode_height;
	int mode_depth;
} videomode;

typedef struct {
    int     writeoffset; /* off set for writing entries to windows */
	int		totalentries; /* total entries read in */
	int		marked;  /* OFF if not ON if yes */
	char	dirpath[128]; /* the current directory path */
	char 	prevdir[128]; /* prevous dirctory */
	}  DIRINFO; /* structure for windows */

struct LINKDIR_PROTO {
    struct LINKDIR_PROTO  *nextentry; /* next entry */
    struct LINKDIR_PROTO  *preventry; /* prevous entry */
    int                   itemlength;  /* length of element's name */
	char	 		      itemtype; /* entry type file or dircetory */
    char                  itemname[128]; /* enrty name */
	int					  itemmarked; /* is it marked for viewing? */
	int					  itemwindow; /* window in which the item occures */
    int					  itemplacement; /* where the item is in the list 0 being the first */
    }; /* LINKDIR prototype */

typedef struct LINKDIR_PROTO LINKDIR; /* define LINKDIR */

struct BEST_MODE {
	struct BEST_MODE	*nextentry;
	struct BEST_MODE	*preventry;
    int 				min_width;
    int 				min_height;
	int					mode;
    int					mode_width;
    int					mode_height;
    int					mode_depth;
    char				*mode_name;
};

typedef struct BEST_MODE bmode; /* define bmode */

extern videomode vmode;
extern bmode     *best_mode_256head;
extern bmode   	 *best_mode_256tail;
extern bmode     *best_mode_32Khead;
extern bmode     *best_mode_32Ktail;
extern bmode     *best_mode_16Mhead;
extern bmode     *best_mode_16Mtail;
extern bmode     *best_mode_temp;
extern LINKDIR   *linked_directory_head; /* the linked list head */
extern LINKDIR   *linked_directory_tail; /* the linked list head */
extern LINKDIR   *linked_directory_temp; /* a linked list element used in presort */
extern LINKDIR   *linked_directory; /* a linked list element */
extern DIRINFO   dirwininfo; /* global struct for windows and position info */
extern opts      options; /* global struct for graphics viewing options */
extern WINDOW    *window[]; /* golbal decloration for the window array */ 
extern FILE      *dfd;

/* functions in frontend.c */
void dorefreshwindows(void); /* refreah the windows */
void dealwithscroll(void); /* deal with input from user */
void start_up_windows(void); /* called from main */
void redrawscreen(void); /* refresh the screen */
void markcurrent(void); /* current element to mark */

/* functions in movement.c */
void moverightwin(void);  /* move right a window */
void moveleftwin(void);  /* move left onw window */
void movedown(void);  /* move down one position */
void moveup(void);  /* move up one position */
void moveend(void); /* move to the begining of the list */
void movehome(void); /* move to the end of the list */
void clear_itemwindow(void); /* clear all itemwindows to -1 */
void mark_window(int ewin, int opts); /* make marked window's text look cool */

/* functions in windows.c */
void runhelp(void); /* pop up the help main window */
void runopthelp(void); /* pop up the help options window */
void enternewdir(char *path_to_change); /* getting path info window and operations */
void runexit(int d); /* exit the program and deallicate memory */
void options_menu(); /* change graphics viewer options */


/* external functions in dirfunct.c */
int  readdirectories(void); /* read directories and make linked list */
void writeoutlist(int writeoffset); /* write the list to windows */
void file_copy();
void file_del();

/* external functions in fileinfo.c */
void file_info(char *file_name, int length);
void getmodeinfo();

/* external functions in viewtext.c */
int read_text(char *file_name);

/* external functions in viewgrph.c */
void initfont();
void writefont(char *s);
void set_custom_palette(int cmap_length, int c_map[3][256]);
void greyscale_colormap(int cmap_length, int c_map[3][256]);
int  read_graphics(char *graphtoread);
int  return_value(int ar[3], int j);
int  service_check(char *read_file);
int  convert_to_int(char c);

/* external functions in scaling.c */
void
scale(int components, int width_old, int height_old,  int width_new, int height_new, int *image_red, int *image_green,
   	  int *image_blue);

/* external functions found in bmp.c */
int  read_bmp_file(char *image_name, int *height, int *width, int *components, int *cmap_switch, int **image_red,
        	      int **image_green, int **image_blue, int cmap[3][256], int *cmap_length);
              
/* external functions found in pixmap.c */
int read_pixmap_file(char *image_name, int *height, int *width, int *components, int **image_red, int **image_green,
            	  int **image_blue, int cmap[3][256], int *cmap_length, int *cmap_switch);
                 
/* external functions found in xbitmap.c */
int read_xbitmap_file(char *image_name, int *height, int *width, int *components, int **image_red, int **image_green, 
                  int **image_blue, int cmap[3][256], int *cmap_length, int *cmap_switch);

/* external functions found in png.c */
int read_png_file(char *image_name, int *height, int *width, int *components, int *cmap_switch, int **image_red,
        	      int **image_green, int **image_blue, int cmap[3][256], int *cmap_length);

/* external functions found in jpeg.c */
int read_jpeg_file(char *image_name, int *height, int *width, int *components, int *cmap_switch, int **image_red,
        	      int **image_green, int **image_blue, int cmap[3][256], int *cmap_length);

/* external functions found in gif.c */
int read_gif_file(char *image_name, int *height, int *width, int *components, int *cmap_switch, int **image_red,
        	      int **image_green, int **image_blue, int cmap[3][256], int *cmap_length);

/* external functions found in tiff.c */
int read_tiff_file(char *image_name, int *height, int *width, int *components, int *cmap_switch, int **image_red,
        	      int **image_green, int **image_blue, int cmap[3][256], int *cmap_length);

/* external functions found in ppm.c */
int read_ppm_file(char *image_name, int *height, int *width, int *components, int *cmap_switch, int **image_red,
        	      int **image_green, int **image_blue, int cmap[3][256], int *cmap_length);
