/* taken from libtiff3.4BETA tool gif2tiff by Sam Leffler   */
/* which was based on based Paul Haeberli's fromgif program */
/* many thanks to Sam Leffler and Paul Haeberli's for clean */
/* readable /understandable code                          */

#include "frontend.h"

#ifndef NO_GIF

#define red_color   0
#define green_color 1
#define blue_color  2

struct image_data {
	int  height; /* image height */
	int  width; /* image width */
	int  cmap_length; /* number of color 0 means ma for mode */
	int  cmap[3][256]; /* color table for 8 bit bmp's */
	int  components; /* components used ie color depth */
	int  cmap_switch;
    int  interlaced;
    char *gif_image;
	FILE *gif;
};

struct image_data image;
unsigned char *stackp;
unsigned int prefix[4096];
unsigned char suffix[4096];
unsigned char stack[4096];
int datasize, codesize, codemask;
int clear, eoi;
int avail, oldcode;

int
process(register int code, unsigned char** fill)
{
    int incode;
    static unsigned char firstchar;

    if (code == clear) {
	codesize = datasize + 1;
	codemask = (1 << codesize) - 1;
	avail = clear + 2;
	oldcode = -1;
	return 1;
    }

    if (oldcode == -1) {
	*(*fill)++ = suffix[code];
	firstchar = oldcode = code;
	return 1;
    }
    if (code > avail) {
	fprintf(dfd, "code %d too large for %d\n", code, avail);
	return 0; 
    }

    incode = code;
    if (code == avail) {      /* the first code is always < avail */
	*stackp++ = firstchar;
	code = oldcode;
    }
    while (code > clear) {
	*stackp++ = suffix[code];
	code = prefix[code];
    }

    *stackp++ = firstchar = suffix[code];
    prefix[avail] = oldcode;
    suffix[avail] = firstchar;
    avail++;

    if (((avail & codemask) == 0) && (avail < 4096)) {
	codesize++;
	codemask += avail;
    }
    oldcode = incode;
    do {
	*(*fill)++ = *--stackp;
    } while (stackp > stack);
    return 1;
}


int
readraster(void) {

    unsigned char *fill = image.gif_image;
    unsigned char buf[255];
    register int bits=0;
    register unsigned long datum=0;
    register unsigned char *c;
    register int count, code;
    int status = 1;

    datasize = getc(image.gif);
    clear = 1 << datasize;
    eoi = clear + 1;
    avail = clear + 2;
    oldcode = -1;
    codesize = datasize + 1;
    codemask = (1 << codesize) - 1;
    for (code = 0; code < clear; code++) {
	prefix[code] = 0;
	suffix[code] = code;
    }
    stackp = stack;
    for (count = getc(image.gif); count > 0; count = getc(image.gif)) {
	fread(buf, 1, count, image.gif);
	for (c = buf; count-- > 0; c++) {
	    datum += (unsigned long) *c << bits;
	    bits += 8;
	    while (bits >= codesize) {
		code = datum & codemask;
		datum >>= codesize;
		bits -= codesize;
		if (code == eoi) goto exitloop;
        if (!process(code, &fill)) {
		    status = 0;
		    goto exitloop;
		}
	    }
	}
	if (fill >= image.gif_image + image.width * image.height) {
	    fprintf(dfd, "raster full before eoi code\n");
	    break;
	}
    }
exitloop:
    if (fill != image.gif_image + (image.width * image.height) ) fprintf(dfd, "warning\n");
    return status;
}

   
int
readgifimage(char* mode) {

    unsigned char buf[9];
    int local;
    unsigned char localmap[256][3];
    int localbits;
    int status, i;

    fread(buf, 1, 9, image.gif);

    image.width      = buf[4] + (buf[5] << 8);
    image.height     = buf[6] + (buf[7] << 8);
    local            = buf[8] & 0x80;
    image.interlaced = buf[8] & 0x40;

    image.gif_image = (char*)malloc(image.width * image.height + 128);

   if (local) {
      localbits = (buf[8] & 0x7) + 1;
      image.cmap_length = (1<<localbits);
fprintf(dfd, "localbits is %d\n", (1<<localbits) );
      for(i = 0; i < (1<<localbits); i++) {
         image.cmap[red_color][i]   = getc(image.gif);
         image.cmap[green_color][i] = getc(image.gif);
         image.cmap[blue_color][i]  = getc(image.gif);
      }
   }

   if (status = readraster()) /* rasterize(interleaved, mode); */
   return status;
}


void
readextension(void)
{
    int count;
    char buf[255];

    (void) getc(image.gif);
    while (count = getc(image.gif)) fread(buf, 1, count, image.gif);
}


int
read_gif_file(char *image_name, int *height, int *width, int *components, int *cmap_switch, int **image_red,
        	   int **image_green, int **image_blue, int cmap[3][256], int *cmap_length) {

   char sig[7], c, *mode = "w";
   int	global, globalbits, i, j, k, array_size;
   
   image.gif = fopen(image_name, "rb");

   fread(sig, 1, 6, image.gif);
fprintf(dfd, "%s\n", sig);

   fread(sig, 1, 7, image.gif);
   global = sig[4] & 0x80;
   if(global) {
      globalbits = (sig[4] & 0x07) + 1;
      image.cmap_length = (1 << globalbits);
fprintf(dfd, "globalbits is %d\n", (1<<globalbits) );
      for(i = 0; i < (1<<globalbits); i++) {
         image.cmap[red_color][i]   = getc(image.gif);
         image.cmap[green_color][i] = getc(image.gif);
         image.cmap[blue_color][i]  = getc(image.gif);
      }
   }
   
   while( (c = getc(image.gif)) != ';' && c != EOF) {
      switch (c) {
         case '\0':  break;
         case ',':   
            readgifimage(mode);
            mode = "a";
            break;
         case '!':   
            readextension();
            break;
      }
   }
   
   if(options.opt_forcemode == 0) image.components = 1;
   else {
		switch (options.opt_forcemode) {
			case 12:
			case 11:
		  	case 10:
		  	case 5:
                image.components = 1;
				break;
            case 14: 
            case 17: 
            case 20: 
            case 23:
                image.components = 2;
                break;
		  	default :
                image.components = 3;
				break;
		}
   }

   array_size   = (int)( (image.width * image.height) * 1.5);
   *image_red   = (int*)malloc(array_size * sizeof(int));
   *image_green = (int*)malloc(array_size * sizeof(int));
   *image_blue  = (int*)malloc(array_size * sizeof(int));

   if(image.interlaced == 0) {
      if(image.components == 1) {
fprintf(dfd, "using low res\n");
         for(i = 0; i < image.width * image.height; (*image_red)[i] = (int)image.gif_image[i++] );
      }
      else {
fprintf(dfd, "using heigh res\n");
         for(i = 0; i < image.width * image.height; i++) {
            (*image_red)[i]   = image.cmap[red_color][ image.gif_image[i] ];
            (*image_green)[i] = image.cmap[green_color][ image.gif_image[i] ];
            (*image_blue)[i]  = image.cmap[blue_color][ image.gif_image[i] ];
         }
      }
   }
   else {
      /* what this does is set up interlacing support with 4 passes the 1st is evry 8th row  */
      /* set from 0, the 2nd is every 8th row, set from 4, the 3rd is every 4th row set from */
      /* 2 and the 4th is every other row starting at 1 */
      if(image.components == 1) {
         for(i = k = 0; i < image.width * image.height; i += image.width * 8) {
            for(j = 0; j < image.width; j++) (*image_red)[i + j]   = (int)image.gif_image[k++];
         }
         for(i = (image.width * 4); i < image.width * image.height; i += image.width * 8) {
            for(j = 0; j < image.width; j++) (*image_red)[i + j]   = (int)image.gif_image[k++];
         }
         for(i = (image.width * 2); i < image.width * image.height; i += image.width * 4) {
            for(j = 0; j < image.width; j++) (*image_red)[i + j]   = (int)image.gif_image[k++];
         }
         for(i = image.width; i < image.width * image.height; i += image.width * 2) {
            for(j = 0; j < image.width; j++) (*image_red)[i + j]   = (int)image.gif_image[k++];
         }
      }
      else {
fprintf(dfd, "using heigh res\n");
         for(i = k = 0; i < image.width * image.height; i += image.width * 8) {
            for(j = 0; j < image.width; j++) {
               (*image_red)[i]   = image.cmap[red_color][ image.gif_image[i] ];
               (*image_green)[i] = image.cmap[green_color][ image.gif_image[i] ];
               (*image_blue)[i]  = image.cmap[blue_color][ image.gif_image[i] ];
            }
         }
         for(i = (image.width * 4); i < image.width * image.height; i += image.width * 8) {
            for(j = 0; j < image.width; j++) {
               (*image_red)[i]   = image.cmap[red_color][ image.gif_image[i] ];
               (*image_green)[i] = image.cmap[green_color][ image.gif_image[i] ];
               (*image_blue)[i]  = image.cmap[blue_color][ image.gif_image[i] ];
            }
         }
         for(i = (image.width * 2); i < image.width * image.height; i += image.width * 4) {
            for(j = 0; j < image.width; j++) {
               (*image_red)[i]   = image.cmap[red_color][ image.gif_image[i] ];
               (*image_green)[i] = image.cmap[green_color][ image.gif_image[i] ];
               (*image_blue)[i]  = image.cmap[blue_color][ image.gif_image[i] ];
            }
         }
         for(i = image.width; i < image.width * image.height; i += image.width * 2) {
            for(j = 0; j < image.width; j++) {
               (*image_red)[i]   = image.cmap[red_color][ image.gif_image[i] ];
               (*image_green)[i] = image.cmap[green_color][ image.gif_image[i] ];
               (*image_blue)[i]  = image.cmap[blue_color][ image.gif_image[i] ];
            }
         }
      }
   }
fprintf(dfd, "finnished with res\n");

   image.cmap_switch = 0;
   if(image.components == 1) {
      for(i = 0; i < image.cmap_length; i++) {
        if(image.cmap[red_color][i] == 0 && image.cmap[green_color][i] == 0 && image.cmap[blue_color][i] == 0) { 
            image.cmap_switch = i;
            break;
        }
      }

      if(i >= image.cmap_length) {
        if(i < 256) { 
           i++;
           image.cmap_length++;
           image.cmap[red_color][i]   = 0; 
           image.cmap[green_color][i] = 0;
           image.cmap[blue_color][i]  = 0;
           image.cmap_switch = i - 1;
        }
      }
   }

   memcpy(cmap, image.cmap, (768 * (sizeof(int))) );

   *width       = image.width;
   *height      = image.height;
   *cmap_length = image.cmap_length;
   *components  = image.components;
   *cmap_switch = image.cmap_switch;
   
   return 1;
}

#endif
