#include "frontend.h"

#ifndef NO_JPEG

#include <jpeglib.h>
#include <setjmp.h>

#define red_color   0
#define green_color 1
#define blue_color  2

struct my_error_mgr {
   struct jpeg_error_mgr pub;
   jmp_buf setjmp_buffer;
};
typedef struct my_error_mgr * my_error_ptr;


METHODDEF void
my_error_exit (j_common_ptr cinfo) {
    my_error_ptr myerr = (my_error_ptr) cinfo->err;
    longjmp(myerr->setjmp_buffer, 1);
}


METHODDEF void
dummy_warning(j_common_ptr cinfo) { 
}

              

int 
read_jpeg_file(char *image_name, int *height, int *width, int *components, int *cmap_switch, int **image_red,
        	   int **image_green, int **image_blue, int cmap[3][256], int *cmap_length) {
   struct     jpeg_decompress_struct cinfo;
   struct     my_error_mgr jerr;
   JSAMPARRAY buffer;
   int	      row_stride, i, j, array_size;
   int 		  lcmap[3][256];
   FILE       *jpeg;   
   
   jpeg = fopen(image_name, "rb");

   cinfo.err               = jpeg_std_error(&jerr.pub);
   jerr.pub.error_exit     = my_error_exit;
   jerr.pub.output_message = dummy_warning;
        
   setjmp(jerr.setjmp_buffer);
   
   jpeg_create_decompress(&cinfo);
    
   jpeg_stdio_src(&cinfo, jpeg);
  
   (void) jpeg_read_header(&cinfo, TRUE);
fprintf(dfd, "read header\n");
    
   if (options.opt_greyscale == 1 || (options.opt_forcemode > 0 && options.opt_forcemode < 13) ) {
       cinfo.desired_number_of_colors = 255;
       if (options.opt_greyscale == 1) {
          cinfo.out_color_space          = JCS_GRAYSCALE;
          cinfo.quantize_colors          = FALSE;
          cinfo.two_pass_quantize        = FALSE;
       }
       else {
          cinfo.quantize_colors          = TRUE;
          cinfo.two_pass_quantize        = TRUE;
       }
   }
fprintf(dfd, "setup components %d\n");
      
   jpeg_start_decompress(&cinfo);
fprintf(dfd, "started decompress\n");

   array_size   = (int)( (cinfo.output_width * cinfo.output_height) * 1.5);
   *image_red   = (int*)malloc(array_size * sizeof(int));
   *image_green = (int*)malloc(array_size * sizeof(int));
   *image_blue  = (int*)malloc(array_size * sizeof(int));

   j = 0;
   row_stride = cinfo.output_width * cinfo.output_components;
   buffer = (*cinfo.mem->alloc_sarray)((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);
   while (cinfo.output_scanline < cinfo.output_height) {
      (void) jpeg_read_scanlines(&cinfo, buffer, 1);

      if(cinfo.output_components == 1) for(i = 0 ; i < cinfo.output_width; (*image_red)[j++] = buffer[0][i++]); 
      else {
         i = 0; 
         while(i < cinfo.output_width * 3) {
            (*image_red)[j]    = buffer[0][i++];
            (*image_green)[j]  = buffer[0][i++];
            (*image_blue)[j++] = buffer[0][i++];
         }
      }
   }

   jpeg_finish_decompress(&cinfo);
fprintf(dfd, "finnished decompress\n");

   if(options.opt_greyscale == 1) {
fprintf(dfd, "using greyscale\n");
      for(i = 0; i < 256; i++) {
         lcmap[red_color][i]   = i;
         lcmap[green_color][i] = i;
         lcmap[blue_color][i]  = i;
      }
   }
   else {
fprintf(dfd, "using color\n");
      for (i = 0; i < cinfo.actual_number_of_colors; i++) {
         lcmap[red_color][i]   = cinfo.colormap[red_color][i];
         lcmap[green_color][i] = cinfo.colormap[green_color][i];
         lcmap[blue_color][i]  = cinfo.colormap[blue_color][i];
      }
   }
fprintf(dfd, "setup cmap\n");

   
   *cmap_switch = 0;
   if(cinfo.output_components == 1) {
      for(i = 0; i < cinfo.actual_number_of_colors; i++) {
        if(lcmap[red_color][i] == 0 && lcmap[green_color][i] == 0 && lcmap[blue_color][i] == 0) { 
            *cmap_switch = i;
            break;
        }
      }

      if(i >= cinfo.actual_number_of_colors) {
        if(i < 256) { 
           i++;
           cinfo.actual_number_of_colors++;
           lcmap[red_color][i]   = 0; 
           lcmap[green_color][i] = 0;
           lcmap[blue_color][i]  = 0;
           *cmap_switch = i - 1;
        }
      }
   }
fprintf(dfd, "setup cmap_switch as %d\n", *cmap_switch);
   
   memcpy(cmap, lcmap, (768 * (sizeof(int))) );

   *width  		= cinfo.output_width;
   *height      = cinfo.output_height;
   *cmap_length = cinfo.actual_number_of_colors;
   *components  = cinfo.output_components;

   jpeg_destroy_decompress(&cinfo);

   fclose(jpeg);
    
   return 1;
}

#endif
