#include "frontend.h"

#ifndef NO_PIXMAP

#include <sys/stat.h>
#include <vgagl.h>
#include <vga.h>

#define red_color   0
#define green_color 1
#define blue_color  2

struct XCOLOR_PROTO {
	struct XCOLOR_PROTO  *nextentry; /* next entry */
	struct XCOLOR_PROTO  *preventry; /* prevous entry */
	int		   			 red; /* amount of red */
	int		 		     green; /* amount of green */
	int		  			 blue; /* amount of blue */
	char		     	 name[128]; /* colors name */
}; /* XCOLOR map prototype */

typedef struct XCOLOR_PROTO XCOLOR; /* define XCOLOR */

XCOLOR *linked_xcolor_head; /* the xcolor list head */
XCOLOR *linked_xcolor_tail; /* the xcolor list tail */
XCOLOR *linked_xcolor_temp; /* a xcolor list element */
XCOLOR *linked_xcolor; /* a xcolor list element */

struct image_data { /* global data */
	char   *pixmap;
    int	   locate;
	int	   width;
	int	   height;
	int	   cmap_length;
	int    cmap[3][256];
	int	   typelength;
	int    cmap_switch;
	int	   components;
};

struct image_data image;


XCOLOR* 
xcolor_preordersort(XCOLOR *head, XCOLOR *tobeadded) {

	/* haed case */
   if (strcmp(tobeadded->name, head->name) < 0) {
   	head->preventry = tobeadded;
   	tobeadded->nextentry = head;
	tobeadded->preventry = NULL;
	return tobeadded;
   }
   else {
	linked_xcolor_temp = head;
	while(strcmp(tobeadded->name, linked_xcolor_temp->name) > 0 && linked_xcolor_temp->nextentry != NULL) {
	   linked_xcolor_temp = linked_xcolor_temp->nextentry; 
	}
	
	/* greater than last entry */
	if (strcmp(tobeadded->name, linked_xcolor_temp->name) == 1) {
	   tobeadded->preventry = linked_xcolor_temp;
	   if(linked_xcolor_temp->nextentry != NULL) { /* not tail */
	     linked_xcolor_temp->nextentry->preventry = tobeadded;
	     tobeadded->nextentry = linked_xcolor_temp->nextentry;
	   }
	   else {
	     tobeadded->nextentry = NULL;
		 linked_xcolor_tail = tobeadded;
	   } 
	   linked_xcolor_temp->nextentry = tobeadded;
       return head;
   	}
	else { /* less than last entry */
	   tobeadded->nextentry = linked_xcolor_temp;
	   linked_xcolor_temp->preventry->nextentry = tobeadded;
	   tobeadded->preventry = linked_xcolor_temp->preventry;
	   linked_xcolor_temp->preventry = tobeadded;
	   return head;
	}
   }
}


void
killcolorlist(XCOLOR *listtype) {

 int i = 0; 

 /* starting with the head and going till tail free up the list */
 if (listtype != NULL) {
   while(listtype->nextentry != NULL) {
		listtype = listtype->nextentry;
		free(listtype->preventry); 
   }
 free(listtype);
 }
}


void
get_xcolor_table() {

	int i = 0, j, v;
	char c, line[128];
	FILE *rgb;

	if((rgb = fopen("/usr/X11/lib/X11/rgb.txt", "r")) == NULL) { 
		fprintf(dfd, "could not open text file for reading\n"); 
		exit(0);
	}

	/* skip fist line */
	for( ; c != '\n'; c = getc(rgb) );
	c = getc(rgb);

	i = 0;
	do {
		linked_xcolor = malloc( sizeof(XCOLOR) );
		linked_xcolor->preventry = NULL;
		linked_xcolor->nextentry = NULL;

		/* read in each line */
		for(j = 0; c != '\n'; c = getc(rgb) ) { line[j++] = c; }
		line[j++] = c;
		line[j] = '\0';
		c = getc(rgb);
		
		linked_xcolor->red = 0;
		for(j = 0; j < 3; j++) {			
			if(line[j] != ' ') {
				if(j == 0) { linked_xcolor->red = ((convert_to_int(line[j]))*100); }
				else if(j == 1) { linked_xcolor->red += ((convert_to_int(line[j]))*10); }
				else { linked_xcolor->red += (convert_to_int(line[j])); }
			}
		}

		for(j = 4; j < 8; j++) {			
			if(line[j] != ' ') {
				if(j == 4) { linked_xcolor->green = ((convert_to_int(line[j]))*100); }
				else if(j == 5) { linked_xcolor->green += ((convert_to_int(line[j]))*10); }
				else { linked_xcolor->green += (convert_to_int(line[j])); }
			}
		}

		for(j = 8; j < 11; j++) {			
			if(line[j] != ' ') {
				if(j == 8) { linked_xcolor->blue = ((convert_to_int(line[j]))*100); }
				else if(j == 9) { linked_xcolor->blue += ((convert_to_int(line[j]))*10); }
				else { linked_xcolor->blue += (convert_to_int(line[j])); }
			}
		}

		v = 0;
		for(j = 13; line[j] != '\n'; j++) {
			if(line[j] != '\t') {
				linked_xcolor->name[v] = line[j];
				v++;
			}
		}

	   	if(i != 0) {
			linked_xcolor_head  = xcolor_preordersort(linked_xcolor_head, linked_xcolor); 
	   	}
	   	else { 
			linked_xcolor_head   = linked_xcolor;
 		}
    	i++;
	} while(c != EOF);
}


void
guess_color_codes(char *three_values,  int j) {

	int i, num[13], remove_value, k;
	char c;

fprintf(dfd, "guessing color\n");
	c = '1';
	remove_value = c - 1;

    if(three_values[1] == '0' && three_values[2] == '0' && three_values[3] == '0') {
       image.cmap[red_color][j] = image.cmap[green_color][j] = image.cmap[blue_color][j]  = 0;
    }
    else {
       for(i = 1; i < 13; i++) {
		  switch(three_values[i]) {
			case '1': case '2': case '3': case '4': case '5': case '6':
			case '7': case '8':	case '9': case '0':
				num[i] = three_values[i] - remove_value;
				break;
			case 'a': case 'A':
				num[i] = 10;
				break;
			case 'b': case 'B':
				num[i] = 11;
				break;
			case 'c': case 'C':
				num[i] = 12;
				break;
			case 'd': case 'D':
				num[i] = 13;
				break;
			case 'e': case 'E':
				num[i] = 14;
				break;
			case 'f': case 'F':
				num[i] = 15;
				break;
		  };
	   }
       for(i = 1, k = 0; i < 13; i++) { if(num[i] >= 0 && num[i] <= 15) { k++; } }
       if(k > 8) {
fprintf(dfd, " using high scan\n");
          image.cmap[red_color][j]   = (num[1] * 16) + num[2];
          image.cmap[green_color][j] = (num[5] * 16) + num[6];
          image.cmap[blue_color][j]  = (num[9] * 16) + num[10];
       }
       else {
fprintf(dfd, " using low scan\n");
          image.cmap[red_color][j]   = (num[1] * 16) + num[2];
          image.cmap[green_color][j] = (num[3] * 16) + num[4];
          image.cmap[blue_color][j]  = (num[5] * 16) + num[6];
       }
       
    }
}


void
check_for_name(char *color_name, int i) {
	
fprintf(dfd, "checking for name\n");
    linked_xcolor = linked_xcolor_head;
	while( strcmp(color_name, linked_xcolor->name) != 0) {
		if(strcmp(linked_xcolor->name, linked_xcolor_tail->name) == 0) break;
		else linked_xcolor = linked_xcolor->nextentry;
	}
    image.cmap[red_color][i]   = linked_xcolor->red;
    image.cmap[green_color][i] = linked_xcolor->green;
	image.cmap[blue_color][i]  = linked_xcolor->blue;
}


void
display_pixmap(int *image_red, int *image_green, int *image_blue) {
   char line[128], name[128], type[256][2], read_string[3];
   char current_type, current_type1, current_string[3];
   int  i, j, k, row, col, current_number;
   
   for(i = 0; i < image.cmap_length; i++) {
      for(image.locate++; image.pixmap[image.locate] != '"'; image.locate++);
      for(j = 0, image.locate++; image.pixmap[image.locate] != '"'; line[j++] = image.pixmap[image.locate++]);
      line[j] = '\0';
fprintf(dfd, "line[%d] is %s\n", i, line);

      if(image.typelength == 1) {
         type[i][0] = line[0];
         j = 2;
      }
      else {
         type[i][0] = line[0];
         type[i][1] = line[1];
         j = 3;
      }
		for(; line[j] != 'c' ; j++); j += 2;
		for(k = 0; line[j] != ' '; name[k++] = line[j++]); name[k] = '\0';
fprintf(dfd, "name is %s\n", name);
		if(name[0] == '#') guess_color_codes(name, i);
		else { 
			if(strcasecmp(name, "none") != 0) { check_for_name(name, i); }
			else { image.cmap[red_color][i] = image.cmap[green_color][i] = image.cmap[blue_color][i] = 255; }
		}
   }
   
   if(image.typelength == 1) {
fprintf(dfd, "using 1bit type\n");
      if(image.components == 1) {
   	  	 for(row = i = j = 0; row < image.height - 1; row++) {
            for(image.locate++; image.pixmap[image.locate] != '"'; image.locate++);
            for(i = 0, image.locate++; image.pixmap[image.locate] != '"'; line[i++] = image.pixmap[image.locate++]);
            line[i] = '\0';
            current_number = 0;
            current_type = type[current_number][0];

         	for(col = 0; col < image.width; col++) {
               if(line[col] == current_type) { image_red[j++] = current_number; }
               else { 
                  for(current_number = 0; line[col] != type[current_number][0]; current_number++);
                  image_red[j++] = current_number;
                  current_type   = type[current_number][0];
               }
            } /* for col */
         } /*  for row */
      } /* if components */
      else {
   	  	 for(row = i = j = 0; row < image.height - 1; row++) {
            for(image.locate++; image.pixmap[image.locate] != '"'; image.locate++);
            for(i = 0, image.locate++; image.pixmap[image.locate] != '"'; line[i++] = image.pixmap[image.locate++]);
            line[i] = '\0';
            current_number = 0;
            current_type = type[current_number][0];

         	for(col = 0; col < image.width; col++) {
               if(line[col] == current_type) { 
                  image_red[j]    = image.cmap[red_color][current_number];;
                  image_green[j]  = image.cmap[green_color][current_number];;
                  image_blue[j++] = image.cmap[blue_color][current_number];;
               }
               else { 
                  for(current_number = 0; line[col] != type[current_number][0]; current_number++);
                  image_red[j]    = image.cmap[red_color][current_number];;
                  image_green[j]  = image.cmap[green_color][current_number];;
                  image_blue[j++] = image.cmap[blue_color][current_number];;
                  current_type    = type[current_number][0];
               }
            } /* for col */
         } /* for row */
      } /* else components */
   } /* if type 1 */
   else { 
fprintf(dfd, "using 2bit type\n");
   	  if(image.components == 1) {
   	  	 for(row = i = j = 0; row < image.height - 1; row++) {
            for(image.locate++; image.pixmap[image.locate] != '"'; image.locate++);
            for(i = 0, image.locate++; image.pixmap[image.locate] != '"'; line[i++] = image.pixmap[image.locate++]);
            line[i] = '\0';
            current_number = 0;
            current_string[0] = type[current_number][0];
            current_string[1] = type[current_number][1];

         	for(col = 0; col < image.width * 2; col += 2) {
               read_string[0] = line[col];
               read_string[1] = line[col + 1];
               if(read_string[0] == current_string[0] && read_string[1] == current_string[1]) { 
                  image_red[j++] = current_number; 
               }
               else { 
                  current_number = -1;
                  while( read_string[0] != current_string[0] || read_string[1] != current_string[1] ) {
                     current_string[0] = type[++current_number][0];
                     current_string[1] = type[current_number][1];
                  }
                  image_red[j++] = current_number;
                  strcpy(current_string, read_string);
               }
            } /* for col */
         } /*  for row */
      } /* if components */
      else {
   	  	 for(row = i = j = 0; row < image.height - 1; row++) {
            for(image.locate++; image.pixmap[image.locate] != '"'; image.locate++);
            for(i = 0, image.locate++; image.pixmap[image.locate] != '"'; line[i++] = image.pixmap[image.locate++]);
            line[i] = '\0';
            current_number = 0;
            current_string[0] = type[current_number][0];
            current_string[1] = type[current_number][1];

         	for(col = 0; col < image.width * 2; col += 2) {
               read_string[0] = line[col];
               read_string[1] = line[col + 1];
               if(read_string[0] == current_string[0] && read_string[1] == current_string[1]) { 
                  image_red[j]    = image.cmap[red_color][current_number];;
                  image_green[j]  = image.cmap[green_color][current_number];;
                  image_blue[j++] = image.cmap[blue_color][current_number];;
               }
               else { 
                  current_number = -1;
                  while( read_string[0] != current_string[0] || read_string[1] != current_string[1] ) {
                     current_string[0] = type[++current_number][0];
                     current_string[1] = type[current_number][1];
                  }
                  image_red[j]    = image.cmap[red_color][current_number];;
                  image_green[j]  = image.cmap[green_color][current_number];;
                  image_blue[j++] = image.cmap[blue_color][current_number];;
                  strcpy(current_string, read_string);
               }
            } /* for col */
         } /*  for row */
      } /* else components */
   } /* else type 2 */
}


int
read_pixmap_file(char *image_name, int *height, int *width, int *components, int **image_red, int **image_green,
            	 int **image_blue, int cmap[3][256], int *cmap_length, int *cmap_switch) {

	int i, j, save_value[3], array_size;
	char line[30], new_line[15];
	int filept, pixsize;
	struct stat fileinfo;
	FILE *pixmap;

    get_xcolor_table();
fprintf(dfd, "got the xcolormap\n");

	pixmap = fopen(image_name, "r");
	filept = fileno(pixmap);
	fstat(filept, &fileinfo);
fprintf(dfd, "filesize is %d\n", fileinfo.st_size);

	pixsize = fileinfo.st_size;
	image.pixmap = (char*)malloc(pixsize * sizeof(char) + 20);
	fread(image.pixmap, sizeof(char), pixsize, pixmap); 
for(i = 0; i < pixsize; fprintf(dfd, "%c", image.pixmap[i++]) );
	fclose(pixmap);

   for(image.locate = 0; image.pixmap[image.locate] != '"'; image.locate++);
fprintf(dfd, "image.locate is %d\n", image.locate);
   for(j = 0, image.locate++; image.pixmap[image.locate] != '"'; line[j++] = image.pixmap[image.locate++]); 
fprintf(dfd, "line is %s\n", line);

    for(i = 0; i < 14; i++) {
       if(line[i] <= '9' && line[i] >= '0') { new_line[i] = line[i]; }
       else { new_line[i] = ':'; }
    }
    j = 0;
fprintf(dfd, "newline is %s\n", new_line);

    /* width */
    for(i = 0; new_line[i] != ':'; i++) {
        save_value[j] = convert_to_int(new_line[i]);
        j++;
    }
    image.width = return_value(save_value, j);
fprintf(dfd, "width is %d, %d\n", image.width, j);
    j = 0;

    /* height */
    for(i = i++; new_line[i] != ':'; i++) {
        save_value[j] = convert_to_int(new_line[i]);
        j++; 
    }
    image.height = return_value(save_value, j);    
fprintf(dfd, "height is %d, %d\n", image.height, j);
    j = 0;
    
    /* cmap_length */
    for(i = i++; new_line[i] != ':'; i++) {
        save_value[j] = convert_to_int(new_line[i]);
        j++;
    }
    image.cmap_length = return_value(save_value, j);
fprintf(dfd, "cmap_length is %d\n", image.cmap_length);
    j = 0;
    
    for(i = i++; new_line[i] != ':'; i++) {
        save_value[j] = convert_to_int(new_line[i]);
        j++;
    }
    image.typelength = return_value(save_value, j);
fprintf(dfd, "typelength is %d\n", image.typelength);
    j = 0;

    array_size   = (int)( (image.width * image.height) * 1.5);
    *image_red   = (int*)malloc(array_size * sizeof(int));
    *image_green = (int*)malloc(array_size * sizeof(int));
    *image_blue  = (int*)malloc(array_size * sizeof(int));

	if(options.opt_forcemode == 0) { image.components = 1; }
	else {
		switch (options.opt_forcemode) {
			case 12:
			case 11:
		  	case 10:
		  	case 5:
                image.components = 1;
				break;
            case 14: 
            case 17: 
            case 20: 
            case 23:
                image.components = 2;
                break;
		  	default :
				image.components = 3;
				break;
		}
	}

	display_pixmap(*image_red, *image_green, *image_blue);
/*
fprintf(dfd, "returned pixmap is\n");
for(i = 0, j = 0; i < image.height; i++) {
   for(array_size = 0; array_size < image.width; array_size++) fprintf(dfd, "%d", *image_red[j++]);
fprintf(dfd, "\n");
}
*/

    for(i = 0; i < image.cmap_length; i++) {
        if(image.cmap[red_color][i] == 0 && image.cmap[green_color][i] == 0 && image.cmap[blue_color][i] == 0) { 
            image.cmap_switch = i;
            break;
        }
    }

    if(i >= image.cmap_length) {
       if(i < 256) { 
          i++;
          image.cmap_length++;
          image.cmap[red_color][i]   = 0; 
          image.cmap[green_color][i] = 0;
          image.cmap[blue_color][i]  = 0;
          image.cmap_switch = i;
       }
       else { image.cmap_switch = 0; }
    }

    *height       = image.height;
    *width        = image.width;
    *components   = image.components;
    *cmap_length  = image.cmap_length;
	*cmap_switch = image.cmap_switch;
    memcpy(cmap, image.cmap, (768 * (sizeof(int))) );

	killcolorlist(linked_xcolor_head); /* clear the list */
	return 1; 
}

#endif
