/* information for png's taken from libpng and other png documents */

#include "frontend.h"

#ifndef NO_PNG

#include <png.h>

#define red_color   0
#define green_color 1
#define blue_color  2


int
read_png_file(char *image_name, int *height, int *width, int *components, int *cmap_switch, int **image_red,
        	  int **image_green, int **image_blue, int cmap[3][256], int *cmap_length) {

   FILE *png;
   png_color_16 backdrop_color = { 0, 0, 0, 0, 0};
   png_struct *png_system;
   png_info   *png_user;
   double screen_gamma = 1.0; /* just a guess as to correct value */
   int i, j, array_size, row, col, lcmap[3][256], block, stg;
   png_bytef *current_row[3000]; /* set max height of a png to 3000 pixels */
   png_bytef *temp_row;

   png = fopen(image_name, "rb");
   
   png_system = malloc(sizeof(png_struct));
   png_user   = malloc(sizeof(png_info));
   setjmp(png_system->jmpbuf);
   
   png_info_init(png_user);
   png_read_init(png_system);
   png_init_io(png_system, png);
   png_read_info(png_system, png_user);

   if(png_user->color_type == PNG_COLOR_TYPE_GRAY && png_user->bit_depth < 8) png_set_expand(png_system);
   else if (png_user->valid & PNG_INFO_tRNS) png_set_expand(png_system);
   
   if(png_user->valid & PNG_INFO_bKGD) {
      png_set_background(png_system, &(png_user->background), PNG_BACKGROUND_GAMMA_FILE, 1, 1.0);
   }
   else png_set_background(png_system, &backdrop_color, PNG_BACKGROUND_GAMMA_SCREEN, 0, 1.0);
fprintf(dfd, "set background\n");
   
   if(png_user->valid & PNG_INFO_gAMA) png_set_gamma(png_system, screen_gamma, png_user->gamma);
   else png_set_gamma(png_system, screen_gamma, 0.45);
fprintf(dfd, "set gamma\n");
   
   if(png_user->bit_depth < 8) png_set_packing(png_system);
fprintf(dfd, "set packed\n");

   if(png_user->color_type == PNG_COLOR_TYPE_GRAY || png_user->color_type == PNG_COLOR_TYPE_GRAY_ALPHA) {
      png_set_gray_to_rgb(png_system);
fprintf(dfd, "set grey fix\n");
   }


   png_read_update_info(png_system, png_user);
fprintf(dfd, "updated info\n");

fprintf(dfd, "width       is %d\n", png_user->width);
fprintf(dfd, "height      is %d\n", png_user->height);
fprintf(dfd, "depth       is %d\n", png_user->bit_depth);
fprintf(dfd, "pixel_depth is %d\n", png_user->pixel_depth);
fprintf(dfd, "cmap_length is %d\n", png_user->num_palette);
fprintf(dfd, "gamma       is %f\n", png_user->gamma);

   /* set memory up to hold image */
   array_size   = (int)( (png_user->width * png_user->height) * 1.5);
   *image_red   = (int*)malloc(array_size * sizeof(int));
   *image_green = (int*)malloc(array_size * sizeof(int));
   *image_blue  = (int*)malloc(array_size * sizeof(int));
   
   if(png_user->pixel_depth <= 8) array_size = png_user->width + 1; /* if low res */
   else array_size = (png_user->width + 1) * 4;
   
   for(i = 0 ; i < png_user->height; i++) current_row[i] = (png_bytef*)malloc(array_size * sizeof(png_bytef) );
   
   png_read_image(png_system, current_row);

   if(png_user->pixel_depth <= 8) { /* if low res */
fprintf(dfd, "using low res\n");
      for (row = j = 0; row < png_user->height; row++) {
         temp_row = current_row[row];
         for(col = 0; col < png_user->width; col++) {
            (*image_red)[j++] = temp_row[col];
         }
      }
   }
   else if(png_user->pixel_depth == 16) {
      for (row = j = 0; row < png_user->height; row++) {
         temp_row = current_row[row];
         for(col = 0; col < png_user->width * 2;) {
            stg = temp_row[col++];
            stg += (temp_row[col++] << 8);
            (*image_red)[j]    = (block & 0x1f) * 8;
            block >>= 5;
            (*image_green)[j]  = (block & 0x1f) * 8;
            block >>= 5;
            (*image_blue)[j++] = (block & 0x1f) * 8;
         }
      }
   }
   else {
fprintf(dfd, "using high res\n");
      for (row = j = 0; row < png_user->height; row++) {
         temp_row = current_row[row];
         for(col = 0; col < png_user->width * 4;) {
            (*image_red)[j]    = temp_row[col++];
            (*image_green)[j]  = temp_row[col++];
            (*image_blue)[j++] = temp_row[col++];
            col++;
         }
      }
   }
   
   for(i = 0 ; i < png_user->height; i++) free(current_row[i]);
   
   if(png_user->pixel_depth <= 8) {
      /* set up cmap */
      for(i = 0; i < png_user->num_palette; i++) {
         lcmap[red_color][i]   = png_user->palette[i].red;
         lcmap[green_color][i] = png_user->palette[i].green;
         lcmap[blue_color][i]  = png_user->palette[i].blue;
      }
   }

   png_read_end(png_system, png_user);

   /* find hi or low res */   
   if(options.opt_forcemode == 0) {
       if(png_user->pixel_depth <= 8) *components = 1;
       else if(png_user->pixel_depth == 16) *components == 2;
       else *components = 3;
   }
   else {
		switch (options.opt_forcemode) {
			case  5 :
			case 10 :
		  	case 11 :
		  	case 12 :
                *components = 1;
				break;
            case 14 :
            case 17 :
            case 20 :
            case 23 :
                *components = 2;
				break;               
		  	default :
				*components = 3;
				break;
		}
   }

   /* find and set the background color */
   *cmap_switch = 0;
   if(*components == 1) {
      for(i = 0; i < png_user->num_palette; i++) {
        if(lcmap[red_color][i] == 0 && lcmap[green_color][i] == 0 && lcmap[blue_color][i] == 0) { 
            *cmap_switch = i;
            break;
        }
      }

      if(i >= png_user->num_palette) {
        if(i < 256) { 
           i++;
           png_user->num_palette++;
           lcmap[red_color][i]   = 0; 
           lcmap[green_color][i] = 0;
           lcmap[blue_color][i]  = 0;
           *cmap_switch = i - 1;
        }
      }
   }
   /* find and set the background color */

   if(*components == 1) memcpy(cmap, lcmap, (768 * (sizeof(int))) ); /* copy over cmap only on low res */

   *cmap_length = png_user->num_palette;
   *width		= png_user->width;
   *height		= png_user->height;
   
fprintf(dfd, "width       is %d\n", *width);
fprintf(dfd, "height      is %d\n", *height);
fprintf(dfd, "components  is %d\n", *components);
fprintf(dfd, "cmap_length is %d\n", *cmap_length);
fprintf(dfd, "cmap_switch is %d\n\n", *cmap_switch);

   png_read_destroy(png_system, png_user, 0);
   
   free(png_user);
   free(png_system);
   fclose(png);

   return 1;
}

#endif
