/* thanks to lug's targa.c, traga.doc and independent jpeg group's    */
/* libjpeg6.0 rdtarga.c  for information on format and implementation */

#include "frontend.h"

#ifndef NO_TARGA

#define red_color   0
#define green_color 1
#define blue_color  2

#define USE_CMAP 1
#define NOT_CMAP 0

#define MAPPED_IMAGE   1
#define RGB_IMAGE      2
#define RGB_RLE_MAPPED 9
#define RGB_RLE_IMAGE  10

#define UCH(x)	((int) (x))
#define GET_2B(array, offset)  ((unsigned int) UCH(array[offset]) + \
			  (((unsigned int) UCH(array[offset+1])) << 8))

struct image_data {
    int  id_number;
	int  height; /* image height */
	int  width; /* image width */
    int  targa_type;
	int  cmap_length; /* number of color 0 means ma for mode */
    int  cmap[3][256]; /* color table for 8 bit bmp's */
    int  cmap_entry_size;
    int  cmap_type;
    int  cmap_start;
	int  components; /* components used ie color depth */
	int  cmap_switch;
    int  targa_pixel_size;
    char *targa_image;
	FILE *targa;
};

struct image_data image;


void
read_24bit(int *image_red, int *image_green, int *image_blue, int cmap_use) {

   int i, c, j, sig[4], array_size, r, g, b, counter, k;
   
   if(image.targa_type == RGB_IMAGE) {
fprintf(dfd, "running 24bit normal %d\n", image.cmap_entry_size);
      for(i = j = 0 ; i < image.width * image.height; i++) {
         if(image.targa_pixel_size < 3) {
            fread(sig, image.cmap_entry_size, 1, image.targa);
            array_size = 0;
            array_size = (sig[1] << 8) | sig[0];
            image_red[j]    = (array_size & 0x7c00) >> 7;
            image_green[j]  = (array_size & 0x03e0) >> 2;
            image_blue[j++] = (array_size & 0x001f) << 3;
         }
         else {
            image_blue[j]   = getc(image.targa);
            image_green[j]  = getc(image.targa);
            image_red[j++]  = getc(image.targa);
            if(image.targa_pixel_size == 4) (void) getc(image.targa);
         }
      } /* for */
   } /* if non RLE */
   else {
fprintf(dfd, "running 8bit rle\n");
      for(i = 0; i < image.width * image.height; ) {
         c = getc(image.targa);
         if(c & 0x80) {
            counter = c & 0x7F;
            fread(sig, image.targa_pixel_size, 1, image.targa);
            if(image.targa_pixel_size < 3) {
               array_size = 0;
               array_size = (sig[1] << 8) | sig[0];
               r = (array_size & 0x7c00) >> 7;
               g = (array_size & 0x03e0) >> 2;
               b = (array_size & 0x001f) << 3;
            }
            for(k = 0; k < counter; k++, i++) {
               image_blue[j]   = b;
               image_green[j]  = g;
               image_red[j++]  = r;
            }
         }
         else {
            counter = c & 0x7F;
            for(k = 0; k < counter; k++, i++) {
               image_blue[j]  = getc(image.targa);
               image_green[j] = getc(image.targa);
               image_red[j++] = getc(image.targa);
            }
         }
      }
   } 
}


void
read_8bit(int *image_red) {

   int i, c, c1, j, k, counter;
   
   if(image.targa_type == MAPPED_IMAGE) {
fprintf(dfd, "running 8bit normal\n");
      for(i = j = 0 ; i < image.width * image.height; i++) image_red[j++] = getc(image.targa);
   }
   else {
fprintf(dfd, "running 8bit rle\n");
      for(i = 0; i < image.width * image.height; ) {
         c = getc(image.targa);
         if(c & 0x80) {
            counter = c & 0x7F;
            c1 = getc(image.targa);
            for(k = 0; k < counter; k++, i++) image_red[j++] = c1;
         }
         else {
            counter = c & 0x7F;
            for(k = 0; k < counter; k++, i++) image_red[j++] = getc(image.targa);
         }
      }
   } 
}


int
read_targa_file(char *image_name, int *height, int *width, int *components, int *cmap_switch, int **image_red,
        	   int **image_green, int **image_blue, int cmap[3][256], int *cmap_length) {


   char sig[18];
   int array_size, total_planes = 0, small_planes = 0, force_total_planes = 0, i;
   
   image.targa = fopen(image_name, "rb");
   
   fread(sig, 18, 1, image.targa);
   
   image.id_number         = sig[0];
   image.cmap_type        = sig[1];
   image.targa_type       = sig[2];
   image.cmap_start       = GET_2B(sig, 3);
   image.cmap_length      = GET_2B(sig, 5);
   image.cmap_entry_size  = sig[7] / 8;
   image.width            = GET_2B(sig, 12);
   image.height           = GET_2B(sig, 14);
   image.targa_pixel_size       = sig[16] / 8;     /* we'll use it directly */
/*   image.image_descriptor = sig[17]; */

   fprintf(dfd, "Image type %d\n", image.targa_type);
   fprintf(dfd, "%s color map\n", (image.cmap_type ? "With" : "Without"));
   fprintf(dfd, "Cmap origin: %d\n", image.cmap_start );
   fprintf(dfd, "Cmap length: %d\n", image.cmap_length );
   fprintf(dfd, "Cmap entry size: %d\n", image.cmap_entry_size );
   fprintf(dfd, "Image size: %dx%d\n", image.width, image.height );
   fprintf(dfd, "Pixel size: %d\n", image.targa_pixel_size );

   if(image.id_number) {
fprintf(dfd, "in if\n");
      while(image.id_number) {
         (void) getc(image.targa);
         image.id_number--;
      }
   }

   array_size   = (int)( (image.width * image.height) * 1.5);
   *image_red   = (int*)malloc(array_size * sizeof(int));
   *image_green = (int*)malloc(array_size * sizeof(int));
   *image_blue  = (int*)malloc(array_size * sizeof(int));

   switch(image.targa_type) {
      case RGB_IMAGE :
      case RGB_RLE_IMAGE :
         total_planes = 1;
         break;
      case MAPPED_IMAGE :
      case RGB_RLE_MAPPED :
         small_planes = 1;
         if(image.cmap_length > 256) force_total_planes = 1;
         break;
   }

   if(image.cmap_type) {
      for(i = 0; i < image.cmap_length; i++) {
         if(image.cmap_entry_size < 3) {
            fread(sig, image.cmap_entry_size, 1, image.targa);
            array_size = 0;
            array_size = (sig[1] << 8) | sig[0];
            cmap[blue_color][i]  = (array_size & 0x7c00) >> 7;
            cmap[green_color][i] = (array_size & 0x03e0) >> 2;
            cmap[red_color][i]   = (array_size & 0x001f) << 3;
         }
         else {
            cmap[blue_color][i]  = getc(image.targa);
            cmap[green_color][i] = getc(image.targa);
            cmap[red_color][i]   = getc(image.targa);
         }
      }
   }
   
   if(total_planes == 1 || force_total_planes == 1) image.components = 3;
   else image.components = 3;
   
   if(total_planes == 1) read_24bit(*image_red, *image_green, *image_blue, NOT_CMAP);
   else if(force_total_planes == 1) read_24bit(*image_red, *image_green, *image_blue, USE_CMAP);
   else read_8bit(*image_red);

   image.cmap_switch = 0;
   if(image.components == 1) {
      for(i = 0; i < image.cmap_length; i++) {
        if(image.cmap[red_color][i] == 0 && image.cmap[green_color][i] == 0 && image.cmap[blue_color][i] == 0) { 
            image.cmap_switch = i;
            break;
        }
      }

      if(i >= image.cmap_length) {
        if(i < 256) { 
           i++;
           image.cmap_length++;
           image.cmap[red_color][i]   = 0; 
           image.cmap[green_color][i] = 0;
           image.cmap[blue_color][i]  = 0;
           image.cmap_switch = i - 1;
        }
      }
   }

   *width       = image.width;
   *height      = image.height;
   *cmap_length = image.cmap_length;
   *components  = image.components;
   *cmap_switch = image.cmap_switch;
   
return 1;
}

#endif
