#include "frontend.h"
#include "keydef.h"
#include <stdlib.h>
#include <string.h>
#include <vgagl.h>
#include <vga.h>

#define red_color   0
#define green_color 1
#define blue_color  2

/* local definition of vmode */
videomode vmode;

/* the conext for the physical, and refresh screens */
GraphicsContext *physicalscreen;
GraphicsContext *backscreen;

/* font definition */
void *font = NULL;

/* defition of screen size */
int screen_size;

/* define string used in writing stuff to screen */
char out[128];
char modeout[128];
char largemode[128] = "Largest possible video mode";
char smallmode[128] = "Smallest possible video mode";

void
display_shutdown()
{
    vga_setmode(TEXT);
}


videomode
find_best_full_mode(int width, int height, int components) {
	
	int mode;

    if (components == 1) {
		best_mode_temp = best_mode_256tail; 
	}
	else if (components == 2) {
		best_mode_temp = best_mode_32Ktail;
	}
	else {
		best_mode_temp = best_mode_16Mtail;
	}
	
	while(best_mode_temp->preventry != NULL &&  (width < best_mode_temp->min_width 
           && height < best_mode_temp-> min_height)) { best_mode_temp = best_mode_temp->preventry;	}

	/* solve the mode 5 error */
	if(best_mode_temp->mode == 5 && (best_mode_temp->mode_width < width ||
	   best_mode_temp->mode_height < height) ) {
       		if(best_mode_temp->nextentry != NULL) best_mode_temp = best_mode_temp->nextentry;
	}

	vmode.mode	 	  = best_mode_temp->mode;
	vmode.mode_width  = best_mode_temp->mode_width;
	vmode.mode_height = best_mode_temp->mode_height;
	vmode.mode_depth  = best_mode_temp->mode_depth;
	return vmode;
}


int
best_mode(int width, int height, int components) {

fprintf(dfd, "components = %d\n", components);
    if (options.opt_forcemode == 0) {
		vmode = find_best_full_mode(width, height, components);
	}
	else {
		/* set what list to use */
		if(components == 1) {
			best_mode_temp = best_mode_256tail;
		}
		else if (components == 2) {
			best_mode_temp = best_mode_32Ktail;
		}
		else {
			best_mode_temp = best_mode_16Mtail;
		}
		/* set what list to use */
	
		/* check for mode */
		while(best_mode_temp->preventry != NULL && best_mode_temp->mode != options.opt_forcemode) {
			best_mode_temp = best_mode_temp->preventry;
		}

		/* if mode finding fails goto find the best else save the mode found */
		if(best_mode_temp == NULL) {
			options.opt_forcemode = 0;
			vmode = find_best_full_mode(width, height, components);
		}
		else {
			vmode.mode		= best_mode_temp->mode;
			vmode.mode_width  = best_mode_temp->mode_width;
			vmode.mode_height = best_mode_temp->mode_height;
			vmode.mode_depth  = best_mode_temp->mode_depth;
		}
	}
    return vmode.mode;
}


/* add reduction to this */
void
display_init(int width, int height, int components, char *graphtoread, int graphics_color)
{
    vga_modeinfo *vgainfo;
    int modetype;
    div_t size;

    modetype = best_mode(width, height, components);

    vgainfo = vga_getmodeinfo(modetype);
    
    vga_setmode(modetype);
  	
	/* set up the space on the screen as backscreen */
	gl_setcontextvgavirtual(modetype);
	backscreen = gl_allocatecontext();
	gl_getcontext(backscreen);

	/* set up the space on the screen as physicalscreen */
	gl_setcontextvga(modetype);
	physicalscreen = gl_allocatecontext();
	gl_getcontext(physicalscreen);

	/* set up the font for this screen size */
	initfont(modetype);

fprintf(dfd, "graphics_color is %d\n", graphics_color);
   if(components == 1) { gl_clearscreen(graphics_color); }
   else {gl_clearscreen(0); }

fprintf(dfd, "cleared screen\n");
fprintf(dfd, "mode depth %d\n", vmode.mode_depth);
    switch (vmode.mode_depth) { /* fix to say actual mode */
		case 256 :
		   	sprintf(modeout, " Mode is %dx%dx256 ", vmode.mode_width, vmode.mode_height);
       	   	sprintf(out, "%s %dx%dx256", graphtoread, width, height);
	   	   	break;
		case 32768 : 
	   		sprintf(modeout, " Mode is %dx%dx32K ", vmode.mode_width, vmode.mode_height);
       		sprintf(out, "%s %dx%dx32K", graphtoread, width, height);
	   		break;
		case 16777216 :
 	   		sprintf(modeout, " Mode is %dx%dx16M ", vmode.mode_width, vmode.mode_height);
       		sprintf(out, "%s %dx%dx16M", graphtoread, width, height);
	   		break;
	}
}


int
setfixdisplay(int addsub)
{
    int  components;

    switch (vmode.mode) {
	case 5 : case 6 : case 7 : case 8 : case 10 : case 11 : case 12 :
	   components = 1;
	   break;
	case 14 : case 17 : case 20 : case 23 :
	   components = 2;
	   break;
	case 33 : case 34 : case 35 :
	   components = 3;
	   break;
    }

	if(addsub == -1) {
		if(components == 1) best_mode_temp = best_mode_256tail;
		else if(components == 2) best_mode_temp = best_mode_32Ktail;
		else best_mode_temp = best_mode_16Mtail;
        
fprintf(dfd, "mode_depth type is %d\n", best_mode_temp->mode_depth);
		if(best_mode_temp != NULL) {
	 		while (best_mode_temp->mode != vmode.mode && best_mode_temp->preventry != NULL) {
				best_mode_temp = best_mode_temp->preventry;
			}
		}

		if(best_mode_temp->preventry != NULL) { 
			vmode.mode = best_mode_temp->preventry->mode; 
			return vmode.mode;
		}
	}
	else {
		if(components == 1){ best_mode_temp = best_mode_256tail; }
		else if(components == 2){ best_mode_temp = best_mode_32Ktail; }
		else { best_mode_temp = best_mode_16Mtail;  }
		if(best_mode_temp != NULL) {
	 		while (best_mode_temp->preventry != NULL && best_mode_temp->mode != vmode.mode) {
				best_mode_temp = best_mode_temp->preventry;
			}
		}
		if(best_mode_temp->nextentry != NULL) { 
			vmode.mode = best_mode_temp->nextentry->mode; 
			return vmode.mode;
        }
	}			
return -1;
}


int
scroll_until_end()
{
    int c, done = 0, offset = 0, i = 0;
	int con_row, con_col;

	initscr();
	keypad(stdscr, TRUE);
     
	/* copy the physical screen to the virtual screen */
	gl_copyscreen(backscreen);

    while (!done) {
	  raw();
	  noecho();
	  c = getch();
	  noraw();
	  echo();
	switch (c) {
	    case KEY_ESC :
			gl_freecontext(backscreen);
			gl_freecontext(physicalscreen);
           	display_shutdown();
            redrawscreen();
		    return ESC_KEY;
		    break;
	    case KEY_ENTER :
		case KEY_N :
		case KEY_n :
			gl_freecontext(backscreen);
			gl_freecontext(physicalscreen);
            display_shutdown();
            redrawscreen();
			return RETURN_KEY;
		    break;
	    case KEY_BACK :
		case KEY_P :
		case KEY_p :
			gl_freecontext(backscreen);
			gl_freecontext(physicalscreen);
            display_shutdown();
            redrawscreen();
			return BACK_KEY;
			break;
	    case KEY_CTRL_X :
			gl_freecontext(backscreen);
			gl_freecontext(physicalscreen);
            display_shutdown();
            redrawscreen();
			return CTRL_X_KEY;
		    break;
	    case KEY_I :
	    case KEY_i :
			gl_setcontext(backscreen);
			gl_copyscreen(physicalscreen);
			gl_setcontext(physicalscreen);
			writefont(out);
		    break;
		case KEY_M :
		case KEY_m :
			gl_setcontext(backscreen);
			gl_copyscreen(physicalscreen);
			gl_setcontext(physicalscreen);
			writefont(modeout);
		    break;
	    case KEY_ADD :
		    i = setfixdisplay(ADD_KEY);
fprintf(dfd, "returned add mode is %d\n", i);
			if(i != -1) { 
				options.opt_forcemode = i; 
				gl_freecontext(backscreen);
				gl_freecontext(physicalscreen);
                return ADD_KEY;
			}
			else {
                gl_setcontext(backscreen);
                gl_copyscreen(physicalscreen);
                gl_setcontext(physicalscreen);
                writefont(largemode);
				done = 0;
			}	
		    break;
	    case KEY_SUB :
		    i = setfixdisplay(SUB_KEY);
fprintf(dfd, "returned sub mode is %d\n", i);
			if(i != -1) { 
				options.opt_forcemode = i;
				gl_freecontext(backscreen);
				gl_freecontext(physicalscreen);
			    return SUB_KEY;
			}
			else {
                gl_setcontext(backscreen);
                gl_copyscreen(physicalscreen);
                gl_setcontext(physicalscreen);
                writefont(smallmode);
				done = 0;
			}	
		    break;
	    case KEY_SPACE :
			gl_freecontext(backscreen);
			gl_freecontext(physicalscreen);
            display_shutdown();
            redrawscreen();
			return SPACE_KEY;
	    case KEY_R :
	    case KEY_r :
			gl_setcontext(backscreen);
			gl_copyscreen(physicalscreen);
			gl_setcontext(physicalscreen);
			done = 0;
			break;
	  default:
		    done = 0;
		    break;
      }
   }
}


void
display_256x32kx16m(int width, int height, int *image_red, int *image_green, int *image_blue,
                	int colormap[3][256], int cmap_length) {

  	int row, col, offset_height, offset_width;
	int blue, green, red, j = 0; /* for use with 24bit displays */

	/* find the offsets for writing to the screen */
	if(height > vmode.mode_height || width > vmode.mode_width) {
		offset_height = 0;
		offset_width = 0;
	}
	else { /* if to large for screen then offsets are 0,0 */
		offset_height = (vmode.mode_height - height)/2;
		offset_width = (vmode.mode_width - width)/2;
    }
fprintf(dfd, "offset_height %d\n", offset_height);
fprintf(dfd, "offset_width  %d\n", offset_width);
fprintf(dfd, "mode depth is %d\n", vmode.mode_depth);

	/* setup the display palette and view bitmap */
	if(vmode.mode_depth == 256) { /* display in 8bit mode */

        if(options.opt_greyscale == 1) { greyscale_colormap(cmap_length, colormap); } 	
    	set_custom_palette(cmap_length, colormap);

fprintf(dfd, "starting image writing\n");
		for(row = 0; row < height; row++) {
			for(col = 0; col < width; col++) {
				gl_setpixel(col+offset_width, row+offset_height, image_red[j++]);
			}
		}
	}
	else { /* display in true color mode */
fprintf(dfd, "writing to screen\n");
		for(row = 0; row < height; row++) {
			for(col = 0; col < width; col++) {
				red   = image_red[j];
				green = image_green[j];
				blue  = image_blue[j++];
				gl_setpixelrgb(col+offset_width, row+offset_height, red, green, blue); 
			}
		}
	}
}


int
service_check(char *read_file) {
	int  i = 0, j, d;
    FILE *graph;
	char c;
	
    graph = fopen(read_file, "r");
	for(j = 0; j < 100; j++) {
		d = fgetc(graph);
		if(d < 129 && d > 0) { i++; }
	}
    fclose(graph);
    
return i;
}


#ifndef NO_SCALE
void
scaling_check(int *width, int *height, int components, int **image_red, int **image_green, int **image_blue) {
   
   int   new_width, new_height, old_width, old_height;
   float i;
   
   old_width  = *width;
   old_height = *height;

   if(old_height > vmode.mode_height || old_width > vmode.mode_width) {
      for(i = .9; (float)(old_height * i) > (float)vmode.mode_height || 
          (float)(old_width * i) > (float)vmode.mode_width; i -= .1);

      new_height = (int)(i * (float)old_height);
      new_width  = (int)(i * (float)old_width);

      scale(components, old_width, old_height, new_width, new_height, *image_red, *image_green, *image_blue);

      *width  = new_width;
      *height = new_height;
   }
   else if( (old_height * 2) < vmode.mode_height && (old_width * 2) < vmode.mode_width) {
      new_height = (int)(1.5 * (float)old_height);
      new_width  = (int)(1.5 * (float)old_width);

      scale(components, old_width, old_height, new_width, new_height, *image_red, *image_green, *image_blue);

      *width  = new_width;
      *height = new_height;
   }
}
#endif


int
read_graphics(char *filename) {

	FILE *graph_file;
    int *image_red, *image_green, *image_blue;
	int c, height, width, components, cmap_switch, complete = 0;
    int cmap[3][256], cmap_length, test;
int nw, nh, *new_image;

fprintf(dfd, "viewgraph top %d, %s\n", complete, filename);
	graph_file = fopen(filename, "rb");
fprintf(dfd, "opened file\n");
	c = getc(graph_file);
fprintf(dfd, "char is %d, '%c'\n", c, c);

	switch (c) {
#ifndef NO_GIF
		case 'G'  :
fprintf(dfd, "found type to be a gif\n");
			complete = read_gif_file(filename, &height, &width, &components, &cmap_switch, &image_red, &image_green,
                       	  &image_blue, cmap, &cmap_length);
			break;
#endif
#ifndef NO_PHOTOCD
		case '1'  : /* this needs to be changed */
fprintf(dfd, "found type to be a photocd\n");
			complete = read_photocd_file(filename, &height, &width, &components, &cmap_switch, &image_red, &image_green,
                       	  &image_blue, cmap, &cmap_length);
			break;
#endif
#ifndef NO_BMP
		case 'B'  :
            fclose(graph_file);
fprintf(dfd, "found type to be a bmp\n");
			complete = read_bmp_file(filename, &height, &width, &components, &cmap_switch, &image_red, &image_green,
                       	  &image_blue, cmap, &cmap_length);
			break;
#endif
#ifndef NO_JPEG
		case 0xff :
fprintf(dfd, "found type to be a jpeg\n");
			complete = read_jpeg_file(filename, &height, &width, &components, &cmap_switch, &image_red, &image_green,
                       	  &image_blue, cmap, &cmap_length);
			break;
#endif
#ifndef NO_TIFF
		case 'I' :
fprintf(dfd, "found type to be a tiff\n");
			complete = read_tiff_file(filename, &height, &width, &components, &cmap_switch, &image_red, &image_green,
                       	  &image_blue, cmap, &cmap_length);
			break;
#endif
#ifndef NO_PPM
        case 'P'  :
fprintf(dfd, "found type to be a ppm\n");
			complete = read_ppm_file(filename, &height, &width, &components, &cmap_switch, &image_red, &image_green,
                       	  &image_blue, cmap, &cmap_length);
			break;           	
#endif
#ifndef NO_TARGA
		case 0x00 :
fprintf(dfd, "found type to be a targa\n");
			complete = read_targa_file(filename, &height, &width, &components, &cmap_switch, &image_red, &image_green,
                       	  &image_blue, cmap, &cmap_length);
        	break;
#endif
#ifndef NO_PNG
		case 0x89 :
        	fclose(graph_file);
fprintf(dfd, "found type to be a png\n");
			complete = read_png_file(filename, &height, &width, &components, &cmap_switch, &image_red, &image_green,
                       	  &image_blue, cmap, &cmap_length);
			break;
#endif
#ifndef NO_PCX
        case '\n' :
fprintf(dfd, "found type to be a pcx\n");
			complete = read_pcx_file(filename, &height, &width, &components, &cmap_switch, &image_red, &image_green,
                       	  &image_blue, cmap, &cmap_length);
			break;
#endif
		case '/'  :
		case '#'  :
			c = getc(graph_file);
			switch(c) {
#ifndef NO_PIXMAP
				case '*' :
                	c = getc(graph_file);
                	c = getc(graph_file);
                    if(c == 'X') {
                       	fclose(graph_file);
fprintf(dfd, "found type to be pixmap\n");
                        complete = read_pixmap_file(filename, &height, &width, &components, &image_red, &image_green,
                                                	&image_blue, cmap, &cmap_length, &cmap_switch);
                    }   
					break;
#endif
#ifndef NO_XBITMAP
				case 'd' :
fprintf(dfd, "found type to be xbitmap\n");
                        complete = read_xbitmap_file(filename, &height, &width, &components, &image_red, &image_green,
                                                	&image_blue, cmap, &cmap_length, &cmap_switch);
					break;
			}
#endif
			break;
	}

    if(complete == 0) {
                 test = service_check(filename);
fprintf(dfd, "doing service check %d\n", test);
#ifndef NO_TEXT
             	 if(test >= 40) read_text(filename);
#endif
    }
    else {
   		vga_disabledriverreport();
        vga_init(); /* set up vga display */


    	display_init(width, height, components, filename, cmap_switch);
#ifndef NO_SCALE
   		if(options.opt_forcescale == ON) {
           scaling_check(&width, &height, components, &image_red, &image_green, &image_blue); 
        }
#endif
        display_256x32kx16m(width, height, image_red, image_green, image_blue, cmap, cmap_length);
        
        free(image_red);
        free(image_green);
        free(image_blue);

fprintf(dfd, "end of graphic\n\n");
        return scroll_until_end();
    }
}


void 
initfont() {
	if(font != NULL) { free(font); }
	font = malloc(256 * 8 * 8 * BYTESPERPIXEL);
	if(vmode.mode_depth == 256) { gl_expandfont(8, 8, 200, gl_font8x8, font); }
	else { gl_expandfont(8, 8, 200, gl_font8x8, font); }
	gl_setfont(8, 8, font);
	gl_setcontextvga(vmode.mode);
	gl_setwritemode(WRITEMODE_MASKED);
}


void 
writefont(char *s) {
	gl_write( ( WIDTH/2-((strlen(s)*8)/2) ) , HEIGHT-12, s);
}

void
greyscale_colormap(int cmap_length, int c_map[3][256]) { /* works to setup the greyscale palette for 8bit images */
    /* new way is to use .299red + .587green + .114blue */
	int i, color_value;
    
    for(i = 0; i < cmap_length; i++) {
       	color_value = ((c_map[red_color][i] * .299) + (c_map[green_color][i] * .587) + (c_map[blue_color][i] * .114));
    	c_map[red_color][i]   = color_value;
        c_map[green_color][i] = color_value;
        c_map[blue_color][i]  = color_value;
    }
	
}



/* set the palette from the colormap */
void 
set_custom_palette(int cmap_length, int c_map[3][256]) {

    char j;
	Palette palette; /* create the palette */
	int i;
    for (i = 0; i < cmap_length; i++) { /* save each color to palette array as a value of 0 to 63 */
		palette.color[i].red = c_map[red_color][i]/4;
		palette.color[i].green = c_map[green_color][i]/4;
		palette.color[i].blue = c_map[blue_color][i]/4;
	}
    gl_setpalette(&palette); /* set the palette to maped colors */
}


int
return_value(int ar[3], int j) {

   int val;

int i;   
for(i = 0; i < j; i++);
   switch(j) {
      case 1:
fprintf(dfd, "case 1\n");
         return ar[0];
         break;
      case 2:
fprintf(dfd, "case 2\n");
         val =  ar[0] * 10;
         val += ar[1];
         return val;
         break;
      case 3:
fprintf(dfd, "case 3\n");
         val =  ar[0] * 100;
         val += ar[1] * 10;
         val += ar[2];
         return val;
         break;
   }
}


int
convert_to_int(char c) {

	switch(c) {
		case 48 : return 0;
		case 49 : return 1;
		case 50 : return 2;
		case 51 : return 3;
		case 52 : return 4;
		case 53 : return 5;
		case 54 : return 6;
		case 55 : return 7;
		case 56 : return 8;
		case 57 : return 9;
	}
}

